# example from https://www.learnpyqt.com/tutorials/multithreading-pyqt-applications-qthreadpool/

from .qt_imports import *

import traceback, sys

class WorkerSignals(QtCore.QObject):
    '''
    Defines the signals available from a running worker thread.
    Supported signals are:
    finished
        No data
    error
        tuple (exctype, value, traceback.format_exc() )
    result
        object data returned from processing, anything
    progress
        int indicating % progress
    '''
    finished = Signal()
    error = Signal(tuple)
    result = Signal(object)
    progress = Signal(int)


class Worker(QtCore.QRunnable):
    '''
    Worker thread
    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.
    :param callback: The function callback to run on this worker thread. Supplied args and
                     kwargs will be passed through to the runner.
    :type callback: function
    :param args: Arguments to pass to the callback function
    :param kwargs: Keywords to pass to the callback function

    '''

    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()

        # Store constructor arguments (re-used for processing)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()
        self.result_cb = None

        # Add the callback to our kwargs
        self.kwargs['progress_callback'] = self.signals.progress

    def set_progress_callback(self, cb):
        if cb is not None:
            self.signals.progress.connect(cb)

    def set_finished_callback(self, cb):
        if cb is not None:
            self.signals.finished.connect(cb)

    def set_result_callback(self, cb):
        # self.signals.result.connect(cb)
        if cb is not None:
            self.result_cb = cb

    @Slot()
    def run(self):
        """
        Initialise the runner function with passed args, kwargs.
        """
        # Retrieve args/kwargs here; and fire processing using them
        try:
            result = self.fn(*self.args, **self.kwargs)
        except:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            # print(" ***** emit result signal ")
            if self.result_cb is not None:
                self.result_cb(result)  # Return the result of the processing
        finally:
            self.signals.finished.emit()  # Done


class ThreadPool(QtCore.QThreadPool):
    def __init(self):
        super(ThreadPool, self).__init__()
        print("Multithreading with maximum %d threads" % self.threadpool.maxThreadCount())

    def set_worker(self, work_function, *args, **kwargs):
        self.worker = Worker(work_function,  *args, **kwargs)  # Any other args, kwargs are passed to the run function

    def set_worker_callbacks(self, progress_cb=None, finished_cb=None, result_cb=None):
        self.worker.set_progress_callback(progress_cb)
        self.worker.set_finished_callback(finished_cb)
        self.worker.set_result_callback(result_cb)

    def start_worker(self):
        self.start(self.worker)




from qimview.utils.qt_imports import QtWidgets
import argparse
import sys
import json

import sys

from qimview.tests_utils.event_recorder import EventRecorder
from qimview.tests_utils.event_player   import EventPlayer
from qimview.tests_utils.qtdump import *
from qimview.image_viewers.qt_image_viewer import QTImageViewer
from qimview.image_viewers.gl_image_viewer import GLImageViewer
from qimview.image_viewers.image_filter_parameters import ImageFilterParameters
from qimview.image_viewers.image_filter_parameters_gui import ImageFilterParametersGui

from qimview.image_readers import gb_image_reader
from typing import Union

# define a Qt window with an OpenGL widget inside it
# class TestWindow(QtGui.QMainWindow):
class ImView(QtWidgets.QMainWindow):
    def __init__(self, events, params: dict) -> None:
        super().__init__()

        record_file = params['record']
        if record_file is not None:
            self.event_recorder = EventRecorder(filename=record_file)
            self.event_recorder.register_widget(id(self), "TestWindow")
        else:
            self.event_recorder = None

        self.main_widget = QtWidgets.QWidget()
        vertical_layout = QtWidgets.QVBoxLayout()
        self.main_widget.setLayout(vertical_layout)

        self.widget: Union[GLImageViewer, QTImageViewer]

        if params['gl']:
            self.widget = GLImageViewer(event_recorder = self.event_recorder)
        else:
            self.widget = QTImageViewer(event_recorder = self.event_recorder)
        if params['timing']:
            self.widget.display_timing = True
        if record_file is not None:
            self.event_recorder.register_widget(id(self.widget), "widget")

        # if params['input_image'] is None:
        #     # Ask for input file
        #     filename =  QtWidgets.QFileDialog.getOpenFileName(caption="imview: Select input image")
        #     params['input_image'] = filename[0]
        # im = gb_image_reader.read(params['input_image'])
        # self.widget.set_image(im)
        # put the window at the screen position (100, 100)
        self.setGeometry(0, 0, self.widget.width(), self.widget.height())
        self.setCentralWidget(self.main_widget)

        self.filter_params = ImageFilterParameters()
        self.filter_params_gui = ImageFilterParametersGui(self.filter_params, name="TestViewer")
        self.filter_params_gui.set_event_recorder(self.event_recorder)

        hor_layout = QtWidgets.QHBoxLayout()
        self.filter_params_gui.add_blackpoint(hor_layout, self.update_image_intensity_event)
        # white point adjustment
        self.filter_params_gui.add_whitepoint(hor_layout, self.update_image_intensity_event)
        # Gamma adjustment
        self.filter_params_gui.add_gamma(hor_layout, self.update_image_intensity_event)
        # G_R adjustment
        self.filter_params_gui.add_g_r(hor_layout, self.update_image_intensity_event)
        # G_B adjustment
        self.filter_params_gui.add_g_b(hor_layout, self.update_image_intensity_event)

        vertical_layout.addLayout(hor_layout)
        vertical_layout.addWidget(self.widget)

        self.show()

        if events != None:
            event_list = events['events']
            # self.event_recorder.register_widget(self.id(), "TestWindow")
            # self.event_recorder.register_widget(self.widget.id(), "widget")
            event_player = EventPlayer()
            event_player.register_widget("TestWindow", self)
            event_player.register_widget("widget", self.widget)
            self.filter_params_gui.register_event_player(event_player)
            event_player.play_events(event_list=event_list)
    
    def set_image(self, params):
        if params['input_image'] is None:
            # Ask for input file
            filename =  QtWidgets.QFileDialog.getOpenFileName(caption="imview: Select input image")
            params['input_image'] = filename[0]
        im = gb_image_reader.read(params['input_image'])
        self.widget.set_image(im)
        self.widget.image_name = params['input_image']

        if params['ref'] is not None:
            im = gb_image_reader.read(params['ref'])
            print(f"im ref {im}")
            if im:
                self.widget.set_image_ref(im)

    def update_image_intensity_event(self):
        self.widget.filter_params.copy_from(self.filter_params)
        # print(f"parameters {self.filter_params}")
        self.widget.viewer_update()

    def event(self, evt):
        if self.event_recorder is not None:
            if evt.type() in RESIZE_EVENTS:
                self.event_recorder.store_event(self, evt)
        return QtWidgets.QMainWindow.event(self, evt)

    def closeEvent(self, event):
        # Save event list
        if self.event_recorder is not None:
            self.event_recorder.save_screen(self.widget)
            self.event_recorder.save_events()
        event.accept()

def main():
    # import numpy for generating random data points
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('input_image', nargs='?', help='input image (if not specified, will open file dialog)')
    parser.add_argument('-ref', default=None,  help='reference image')
    parser.add_argument('-p', '--play', help='events json file', default=None)
    parser.add_argument('-r', '--record', help='record events in given json file', default=None)
    parser.add_argument('--timing', action='store_true', help='display timings')
    parser.add_argument('--gl', action='store_true', help='use opengl viewer')
    args = parser.parse_args()
    _params = vars(args)
    print(f"record {_params['record']}")


    # create the Qt App and window
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationDisplayName('imview ' + ' '.join(sys.argv[1:]))
    if _params['play'] is not None:
        with open(_params['play']) as json_file:
            events = json.load(json_file)
    else:
        events = None

    window = ImView(events, _params)
    window.set_image(_params)
    window.show()
    app.exec()

if __name__ == '__main__':
    main()



# voir lien interessant https://github.com/geoffsalmon/vlc-python/blob/master/generated/vlc.py

#! /usr/bin/python

#
# Qt example for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#

import sys
import os.path
import vlc
from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtWidgets import QWidget, QFrame
from ..parameters.numeric_parameter import  NumericParameter
from ..parameters.numeric_parameter_gui import NumericParameterGui


unicode = str  # Python 3

# TODO: deal with macOS
# if sys.platform == "darwin":  # for MacOS
#     self.videoframe = QtGui.QMacCocoaViewContainer(0)

if sys.platform == "darwin":  # for MacOS
    BaseWidget = QtGui.QMacCocoaViewContainer
else:
    BaseWidget = QFrame


class myVideoWidget(BaseWidget):
    def __init__(self):
        if sys.platform == "darwin":  # for MacOS
            super().__init__(0)
        else:
            super().__init__()
        self.current_scale = 1
        self.media_player = None

    def new_scale(self, mouse_zy, height):
        return max(1, self.current_scale * (1 + mouse_zy * 5.0 / height))
        # return max(1, self.current_scale  + mouse_zy * 5.0 / height)

    def set_media_player(self, media_player):
        self.media_player = media_player

        # now if we set both inputs mouse and keyboard to false, we are able to catch event from the player!
        self.media_player.video_set_mouse_input(False)
        self.media_player.video_set_key_input(False)

        if sys.platform.startswith('linux'): # for Linux using the X Server
            media_player.set_xwindow(self.winId())
        elif sys.platform == "win32": # for Windows
            media_player.set_hwnd(self.winId())
        elif sys.platform == "darwin": # for MacOS
            media_player.set_nsobject(self.winId())

    def wheelEvent(self, event):
        if hasattr(event, 'delta'):
            delta = event.delta()
        else:
            delta = event.angleDelta().y()
        coeff = delta/5
        rect = self.geometry()
        prev_scale = self.current_scale
        self.current_scale = self.new_scale(coeff, rect.height())
        # print(f" geometry {rect.x()} {rect.y()} {rect.width()} {rect.height()}")
        # print(f" current_scale {self.current_scale}")
        rect.setWidth (rect.width() *(self.current_scale/prev_scale))
        rect.setHeight(rect.height()*(self.current_scale/prev_scale))
        # self.setGeometry(rect)
        # self.setFixedSize(rect.width(), rect.height())

        if self.media_player is not None:
            (video_w, video_h) = self.media_player.video_get_size()
            print(f"media size {video_w}x{video_h}")
            print(f" geometry  {rect.width()}x{rect.height()}")
            print(f" scale approx  {rect.width()/video_w:0.2f}x{rect.height()/video_h:0.2f}")

            geom = self.media_player.video_get_crop_geometry()
            print(f"geom: {geom}")
            print(f"video_get_scale: {self.media_player.video_get_scale()}")
            center_x = video_w/2
            center_y = video_h/2
            new_w = int(video_w/self.current_scale+0.5)
            new_h = int(video_h/self.current_scale+0.5)
            start_x = int((video_w-new_w)/2+0.5)
            start_y = int((video_h-new_h)/2+0.5)
            # not clear why it requires to multiply by 2 the top-left position
            self.media_player.video_set_crop_geometry(f"{new_w}x{new_h}+{start_x*2}+{start_y*2}")
            # else:
            #     self.mediaplayer.video_set_crop_geometry(None)
            # print(f"geom: {self.mediaplayer.video_get_crop_geometry()}")



    def resizeEvent(self, event):
        """Called upon window resizing: reinitialize the viewport.
        """
        # print("resize {} {}  self {} {}".format(event.size().width(), event.size().height(),
        #       self.width(), self.height()))
        # event.ignore()


class VLCPlayer(QWidget):
    """A simple Media Player using VLC and Qt
    """
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Media Player")

        self.play_position = NumericParameter()
        self.play_position.range = [0, 1000]
        self.play_position.float_scale = 1000

        # creating a basic vlc instance
        self.instance = vlc.Instance()
        # creating an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()
        self.synchronize_viewer = None

        self.createUI()
        self.isPaused = False

    def createUI(self):
        """Set up the user interface, signals & slots
        """
        # In this widget, the video will be drawn
        if sys.platform == "darwin": # for MacOS
            self.videoframe = QtGui.QMacCocoaViewContainer(0)
        else:
            self.videoframe = myVideoWidget()
        self.palette = self.videoframe.palette()
        self.palette.setColor (QtGui.QPalette.Window,
                               QtGui.QColor(0,0,0))
        self.videoframe.setPalette(self.palette)
        self.videoframe.setAutoFillBackground(True)

        # create slider
        self.play_position_gui = NumericParameterGui(name="play_position", param=self.play_position,
                                                     callback=lambda: self.synchronize_set_play_position(self))
        self.play_position_gui.decimals = 3

        # self.positionslider = QSlider(QtCore.Qt.Horizontal, self)
        # self.positionslider.setToolTip("Position")
        # self.positionslider.setMaximum(1000)
        # self.positionslider.sliderMoved.connect(self.setPosition)

        self.hbuttonbox = QtWidgets.QHBoxLayout()
        self.playbutton = QtWidgets.QPushButton("Play")
        self.hbuttonbox.addWidget(self.playbutton)
        self.playbutton.clicked.connect(lambda: self.synchronize_toggle_play(self))

        self.stopbutton = QtWidgets.QPushButton("Stop")
        self.hbuttonbox.addWidget(self.stopbutton)
        self.stopbutton.clicked.connect( self.Stop)

        self.hbuttonbox.addStretch(1)
        self.volumeslider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.volumeslider.setMaximum(100)
        self.volumeslider.setValue(self.mediaplayer.audio_get_volume())
        self.volumeslider.setToolTip("Volume")
        self.hbuttonbox.addWidget(self.volumeslider)
        self.volumeslider.valueChanged.connect( self.setVolume)

        self.vboxlayout = QtWidgets.QVBoxLayout()
        self.vboxlayout.addWidget(self.videoframe)
        self.hboxlayout = QtWidgets.QHBoxLayout()
        # self.play_position_gui.create(moved_callback=True)
        self.play_position_gui.add_to_layout(self.hboxlayout)
        self.vboxlayout.addLayout(self.hboxlayout)
        self.vboxlayout.addLayout(self.hbuttonbox)

        self.setLayout(self.vboxlayout)

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(200)
        self.timer.timeout.connect(self.updateUI)

    def set_synchronize(self, viewer):
        self.synchronize_viewer = viewer

    def synchronize_toggle_play(self, event_viewer):
        self.PlayPause()
        if self.synchronize_viewer is not None and self.synchronize_viewer is not event_viewer:
            self.synchronize_viewer.synchronize_toggle_play(event_viewer)

    def PlayPause(self):
        """Toggle play/pause status
        """
        if self.mediaplayer.is_playing():
            self.mediaplayer.pause()
            self.playbutton.setText("Play")
            self.isPaused = True
            # geom = self.mediaplayer.video_get_crop_geometry()
            # if geom is None:
            #     self.mediaplayer.video_set_crop_geometry("520x520+10+10")
            # else:
            #     self.mediaplayer.video_set_crop_geometry(None)
            # print(f"geom: {self.mediaplayer.video_get_crop_geometry()}")
        else:
            if self.mediaplayer.play() == -1:
                self.OpenFile()
                return
            self.mediaplayer.play()
            self.playbutton.setText("Pause")
            self.timer.start()
            self.isPaused = False

    def Stop(self):
        """Stop player
        """
        self.mediaplayer.stop()
        self.playbutton.setText("Play")

    def set_video(self, filename=None):
        self.OpenFile(filename)

    def OpenFile(self, filename=None):
        """Open a media file in a MediaPlayer
        """
        print(f"filename is {filename}")
        if filename is None or not os.path.isfile(filename):
            filename = QtWidgets.QFileDialog.getOpenFileName(self, "Open File", os.path.expanduser('~'))
            if not filename:
                return
            print(f"filename is {filename}")
            filename = filename[0]
            print(f"filename is {filename}")
        if not filename:
            return


        self.media = self.instance.media_new(filename)
        # put the media in the media player
        self.mediaplayer.set_media(self.media)

        # parse the metadata of the file
        self.media.parse()
        # set the title of the track as window title
        self.setWindowTitle(self.media.get_meta(0))

        # the media player has to be 'connected' to the QFrame
        # (otherwise a video would be displayed in it's own window)
        # this is platform specific!
        # you have to give the id of the QFrame (or similar object) to
        # vlc, different platforms have different functions for this

        self.videoframe.set_media_player(self.mediaplayer)

        self.PlayPause()

    def setVolume(self, Volume):
        """Set the volume
        """
        self.mediaplayer.audio_set_volume(Volume)

    def synchronize_set_play_position(self, event_viewer):
        if not self.mediaplayer.is_playing():
            print(f"synchronize_set_play_position {id(self)}")
            self.set_play_position()
            if self.synchronize_viewer is not None and self.synchronize_viewer is not event_viewer:
                self.synchronize_viewer.play_position.copy_from(self.play_position)
                self.synchronize_viewer.synchronize_set_play_position(event_viewer)

    def set_play_position(self):
        print(f"set_play_position {id(self)} {self.play_position.float}")
        self.mediaplayer.set_position(self.play_position.float)

    # def setPosition(self, position):
    #     """Set the position
    #     """
    #     # setting the position to where the slider was dragged
    #     self.mediaplayer.set_position(position / 1000.0)
    #     # the vlc MediaPlayer needs a float value between 0 and 1, Qt
    #     # uses integer variables, so you need a factor; the higher the
    #     # factor, the more precise are the results
    #     # (1000 should be enough)

    def updateUI(self):
        """updates the user interface"""
        # setting the slider to the desired position
        # self.positionslider.setValue(self.mediaplayer.get_position() * 1000)
        self.play_position.float = self.mediaplayer.get_position()
        self.play_position_gui.setValue(self.play_position.int)

        if not self.mediaplayer.is_playing():
            # no need to call this function if nothing is played
            self.timer.stop()
            if not self.isPaused:
                # after the video finished, the play button stills shows
                # "Pause", not the desired behavior of a media player
                # this will fix it
                self.Stop()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    player = VLCPlayer()
    player.show()
    player.resize(640, 480)
    if sys.argv[1:]:
        player.OpenFile(sys.argv[1])
    sys.exit(app.exec_())



#
#
#

import traceback
import inspect
from abc    import abstractmethod
from typing import Optional, Tuple, Callable
import cv2
import numpy as np
from qimview.utils.viewer_image import ViewerImage, ImageFormat
from qimview.image_viewers.image_filter_parameters import ImageFilterParameters
from qimview.utils.utils        import get_time
from qimview.utils.qt_imports   import QtGui, QtCore, QtWidgets
from .fullscreen_helper         import FullScreenHelper
from .image_viewer_key_events   import ImageViewerKeyEvents
from .image_viewer_mouse_events import ImageViewerMouseEvents

QtKeys  = QtCore.Qt.Key
QtMouse = QtCore.Qt.MouseButton

try:
    import qimview_cpp
except ImportError as import_excep:
    HAS_CPPBIND = False
    print(f"Failed to load qimview_cpp: {import_excep}")
else:
    HAS_CPPBIND = True
print(f"Do we have cpp binding ? {HAS_CPPBIND}")

# Copied from:
# https://stackoverflow.com/questions/17065086/how-to-get-the-caller-class-name-inside-
# a-function-of-another-class-in-python
def get_class_from_frame(fr):
    args, _, _, value_dict = inspect.getargvalues(fr)
    # we check the first parameter for the frame function is
    # named 'self'
    if len(args) and args[0] == 'self':
        # in that case, 'self' will be referenced in value_dict
        instance = value_dict.get('self', None)
        if instance:
            # return its class name
            try:
                # return getattr(instance, '__class__', None)
                return getattr(instance, '__class__', None).__name__
            except AttributeError as _e:
                print(f"get_class_from_frame error {_e}")
                return None
    # return None otherwise
    return None


def get_function_name() -> str:
    """ Return function name from traceback """
    return traceback.extract_stack(None, 2)[0][2]

class trace_method():
    """ """
    def __init__(self, tab):
        self.tab = tab
        method = traceback.extract_stack(None, 2)[0][2]
        print(self.tab[0] + method)
        self.tab[0] += '  '

    def __del__(self):
        self.tab[0] = self.tab[0][:-2]


#  @dataclass(slots=True)
class ImageViewer:
    """ Image Viewer base class, holds a member _widget of the inherited class """
    # Use slots to avoid new member creation
    # __slots__ = (
    #     '_widget',
    #     '_width',
    #     '_height',
    #     '_image_name',
    #     '_active',
    #     '_on_active',
    #     '_display_timing',
    #     '_verbose',
    #     '_image',
    #     '_image_ref',
    #     '_histo_ref',
    #     '_histo_rect',
    #     '_histo_scale',
    #     '_fullscreen',
    #     '_on_synchronize',
    #     '_save_image_clipboard',
    #     '_clipboard',
    #     '_key_events',
    #     '_mouse_events',
    #     'data',
    #     'lastPos',
    #     'current_dx',
    #     'current_dy',
    #     'current_scale',
    #     'tab',
    #     'trace_calls',
    #     'filter_params',
    #     'start_time',
    #     'timings',
    #     'replacing_widget',
    #     'before_max_parent',
    #     'show_histogram',
    #     'show_cursor',
    #     'show_overlay',
    #     'show_stats',
    #     'show_image_differences',
    #     'show_intensity_line',
    #     'antialiasing',
    #     'image_id',
    #     'image_ref_id',
    #     'evt_width',
    #     'evt_height',
    #     )

    def __init__(self, widget : QtWidgets.QWidget):
        self._widget          : QtWidgets.QWidget = widget

        # --- Protected members
        self._width           : int = 500
        self._height          : int = 500
        self._image_name      : str = ""
        self._image_ref_name  : str = ""
        # If the window is active, its appearance (text color or more) will be different
        self._active          : bool = False
        self._on_active       : Optional[Callable] = None
        self._display_timing  : bool = False
        self._verbose         : bool = False
        self._image           : Optional[ViewerImage] = None
        self._image_ref       : Optional[ViewerImage] = None
        # Rectangle in which the histogram is displayed
        self._histo_rect      : Optional[QtCore.QRect] = None
        # Histogram displayed scale
        self._histo_scale     : int = 1
        # Area of the text
        self._text_rect : Optional[QtCore.QRect] = None

        # FullScreen helper features
        self._fullscreen      : FullScreenHelper = FullScreenHelper()

        # Synchronization callback
        self._on_synchronize  : Optional[Callable]= None

        # Clipboard
        self._save_image_clipboard : bool                       = False
        self._clipboard            : Optional[QtGui.QClipboard] = None
        # Key event class
        self._key_events   : ImageViewerKeyEvents   = ImageViewerKeyEvents(self)
        # Mouse event class
        self._mouse_events : ImageViewerMouseEvents = ImageViewerMouseEvents(self)
        # Add Mouse help tab
        self._key_events.add_help_tab("ImageViewer Mouse", self._mouse_events.markdown_help())

        # Current mouse information: position, displacement
        self.mouse_displ      : QtCore.QPoint = QtCore.QPoint(0,0)
        self.mouse_pos        : QtCore.QPoint = QtCore.QPoint(0,0)
        self.mouse_zoom_displ : QtCore.QPoint = QtCore.QPoint(0,0)

        # --- Public members
        self.data = None
        self.current_dx = self.current_dy = 0
        self.current_scale = 1
        self.tab = ["--"]
        self.trace_calls           : bool                       = False
        self.filter_params         : ImageFilterParameters      = ImageFilterParameters()

        self.start_time = dict()
        self.timings = dict()
        self.replacing_widget = None
        self.before_max_parent = None
        self.show_histogram         : bool = True
        self.show_cursor            : bool = False
        self.show_overlay           : bool = False
        self.show_stats             : bool = False
        self.show_image_differences : bool = False
        self.show_intensity_line    : bool = False
        self.antialiasing           : bool = True
        # We track an image counter, changed by set_image, to help reducing same calculations
        self.image_id       = -1
        self.image_ref_id   = -1
        # Widget dimensions to be defined in child resize event
        self.evt_width : int
        self.evt_height : int

    # === Properties

    # --- widget
    @property
    def widget(self) -> QtWidgets.QWidget:
        return self._widget

    # --- display_timing
    @property
    def display_timing(self):
        return self._display_timing

    @display_timing.setter
    def display_timing(self, v):
        self._display_timing = v

    # --- verbose
    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, v):
        self._verbose = v

    # --- is_active
    @property
    def is_active(self) -> bool:
        return self._active

    @is_active.setter
    def is_active(self, active: bool =True):
        # Do something only if the active flag changes
        if active != self._active:
            self._active = active

    def activate(self):
        """
            Activate this window and call possible callback if it was not already active
        """
        if self.is_active == False:
            # If a callback is set, rely on it
            self._active = True
            if self._on_active:
                self._on_active(self)

    # --- image_name
    @property
    def image_name(self) -> str:
        return self._image_name

    @image_name.setter
    def image_name(self, v : str):
        self._image_name = v

    # --- image_ref_name
    @property
    def image_ref_name(self) -> str:
        return self._image_ref_name

    @image_ref_name.setter
    def image_ref_name(self, v : str):
        self._image_ref_name = v

    # === Public methods
    def add_help_tab(self, title:str,  help:str) -> None:
        """ Add markdown formatted text to generated help in help dialog """
        self._key_events.add_help_tab(title, help)

    def add_help_links(self, help:str) -> None:
        """ Add markdown formatted text to generated help in help dialog """
        self._key_events.add_help_links(help)

    def set_activation_callback(self, cb : Optional[Callable]) -> None:
        """ Set the activation callback """
        self._on_active = cb

    def set_synchronization_callback(self, cb : Optional[Callable]) -> None:
        """ Set the synchronization callback """
        self._on_synchronize = cb

    def get_image(self) -> Optional[ViewerImage]:
        """ Return the current image """
        return self._image

    def set_image(self, image : Optional[ViewerImage]):
        """ Set the viewer image """
        is_different = (self._image is None) or (self._image is not image)
        if image is not None:
            self.print_log(f'set_image({image.data.shape}): is_different = {is_different}')
        if is_different:
            self._image = image
            self.image_id += 1
        return is_different

    def set_image_ref(self, image_ref : Optional[ViewerImage] = None):
        """ Set the reference image """
        is_different = (self._image_ref is None) or (self._image_ref is not image_ref)
        if is_different:
            self._image_ref = image_ref
            self.image_ref_id += 1

    def set_clipboard(self, clipboard : Optional[QtGui.QClipboard], save_image: bool):
        self._clipboard            = clipboard
        self._save_image_clipboard = save_image

    def print_log(self, mess, force=False):
        if self.verbose or force:
            caller_name = inspect.stack()[1][3]
            print("{}{}: {}".format(self.tab[0], caller_name, mess))

    def start_timing(self, title=None):
        if not self.display_timing: return
        if title is None:
            # it seems that inspect is slow
            caller_name = inspect.stack()[1][3]
            class_name = get_class_from_frame(inspect.stack()[1][0])
            if class_name is not None:
                caller_name = "{}.{}".format(class_name, caller_name)
        else:
            caller_name = title
        self.start_time[caller_name] = get_time()
        self.timings[caller_name] = ''

    def add_time(self, mess, current_start, force=False, title=None):
        if not self.display_timing: return
        if self.display_timing or force:
            if title is None:
                caller_name = inspect.stack()[1][3]
                class_name = get_class_from_frame(inspect.stack()[1][0])
                if class_name is not None:
                    caller_name = "{}.{}".format(class_name, caller_name)
            else:
                caller_name = title
            if caller_name in self.start_time:
                total_start = self.start_time[caller_name]
                ctime = get_time()
                mess = "{} {:0.1f} ms, total {:0.1f} ms".format(mess, (ctime -current_start)*1000, (ctime-total_start)*1000)
                self.timings[caller_name] += "{}{}: {}\n".format(self.tab[0], caller_name, mess)

    def print_timing(self, add_total=False, force=False, title=None):
        if not self.display_timing: return
        if title is None:
            caller_name = inspect.stack()[1][3]
            class_name = get_class_from_frame(inspect.stack()[1][0])
            if class_name is not None:
                caller_name = "{}.{}".format(class_name, caller_name)
        else:
            caller_name = title
        if add_total:
            self.add_time("total", self.start_time[caller_name], force)
        if self.timings[caller_name] != '':
            print(self.timings[caller_name])

    # Note: 'ImageViewer' is a forward reference to ImageViewer class
    def synchronize_data(self, dest_viewer: 'ImageViewer') -> None:
        """ Synchronize: copy parameters to another viewer
        """
        dest_viewer.current_scale = self.current_scale
        dest_viewer.current_dx = self.current_dx
        dest_viewer.current_dy = self.current_dy
        dest_viewer.mouse_displ      = self.mouse_displ
        dest_viewer.mouse_pos        = self.mouse_pos
        dest_viewer.mouse_zoom_displ = self.mouse_zoom_displ

        dest_viewer.show_histogram      = self.show_histogram
        dest_viewer.show_cursor         = self.show_cursor
        dest_viewer.show_intensity_line = self.show_intensity_line
        dest_viewer._histo_scale        = self._histo_scale

    def synchronize(self):
        """ Calls synchronization callback if available
        """
        start_time = get_time() if self.display_timing else None
        if self._on_synchronize: 
            self._on_synchronize(self)
        if start_time:
            print('       End sync --- {:0.1f} ms'.format((get_time()-start_time)*1000))

    def new_scale(self, mouse_zy, height):
        return max(1, self.current_scale * (1 + mouse_zy * 5.0 / self._height))
        # return max(1, self.current_scale  + mouse_zy * 5.0 / height)

    def new_translation(self):
        dx = self.current_dx + self.mouse_displ.x()/self.current_scale
        dy = self.current_dy - self.mouse_displ.y()/self.current_scale
        return dx, dy

    def check_translation(self):
        return self.new_translation()

    @abstractmethod
    def viewer_update(self):
        pass

    def key_press_event(self, event, wsize):
        self._key_events.key_press_event(event, wsize)

    def display_message(self, im_pos: Optional[Tuple[int,int]], scale = None) -> str:
        text : str = f'im:{self.image_name}'
        text += f'\nref:{self.image_ref_name}'
        if self.show_cursor and im_pos and self._image:
            text +=  f"\n {self._image.data.shape} {self._image.data.dtype} prec:{self._image.precision}"
            if scale is not None:
                text += f"\n x{scale:0.2f}"
            im_x, im_y = im_pos
            values = self._image.data[im_y, im_x]
            text += f"\n pos {im_x:4}, {im_y:4} \n {self._image.channels.name[3:]}:{str(values).strip('[]')}"

        # ref_txt = self.image_ref_name if self.image_ref_name else 'ref'
        ref_txt = 'ref'
        if self.show_overlay:
            text += f"\n {ref_txt} | im"
        if self.show_image_differences:
            text += f"\n im - {ref_txt}"
        return text

    def display_text(self, painter: QtGui.QPainter, text: str, font_size=-1) -> None:
        self.start_timing()
        color = QtGui.QColor(255, 50, 50, 255) if self.is_active else QtGui.QColor(50, 50, 255, 255)
        painter.setPen(color)
        # Compute font size base on widget size
        if font_size == -1:
            # from 4 tp 14
            # for now very ad-hoc formula
            s = min(self._widget.width(), self._widget.height())
            font_size = min(14, int(4 + s/130 +0.5))
        font = QtGui.QFont('Decorative', font_size)
        # font.setBold(True)
        painter.setFont(font)
        painter.setBackground(QtGui.QColor(250, 250, 250, int(0.75*255)))
        painter.setBackgroundMode(QtGui.Qt.BGMode.OpaqueMode)
        text_options = \
            QtCore.Qt.AlignmentFlag.AlignTop  | \
            QtCore.Qt.AlignmentFlag.AlignLeft | \
            QtCore.Qt.TextFlag.TextWordWrap
        area_width = 400
        area_height = 200
        # boundingRect is interesting but slow to be called at each display
        # bounding_rect = painter.boundingRect(0, 0, area_width, area_height, text_options, self.display_message)
        margin_x = 8
        margin_y = 5
        painter.drawText(
            margin_x, 
            # self.evt_height-margin_y-bounding_rect.height(), area_width, area_height,
            margin_y, area_width, area_height,
            text_options,
            text
            )
        self._text_rect = painter.boundingRect(margin_x, margin_y, 
                                               area_width, area_height, text_options, text)
        self.print_timing()

    def compute_histogram(self, current_image, show_timings=False):
        # print(f"compute_histogram show_timings {show_timings}")
        h_start = get_time() if show_timings else None
        # Compute steps based on input image resolution
        im_w, im_h = current_image.shape[1], current_image.shape[0]
        target_w = 800
        target_h = 600
        hist_x_step = max(1, int(im_w/target_w+0.5))
        hist_y_step = max(1, int(im_h/target_h+0.5))
        input_image = current_image
        # print(f"current_image {current_image.shape} _image {self._image.shape}")
        # input_image = self._image
        resized_im = input_image[::hist_y_step, ::hist_x_step, :]
        resized_im = input_image
        if self.verbose:
            print(f"qtImageViewer.compute_histograph() steps are {hist_x_step, hist_y_step} "
                f"shape {current_image.shape} --> {resized_im.shape}")
        resized_time = get_time()-h_start if h_start else None

        calc_hist_time = 0

        # First compute all histograms
        start_hist = get_time() if show_timings else None
        hist_all = np.empty((3, 256), dtype=np.float32)
        # print(f"{resized_im[::100,::100,:]}")
        for channel, im_ch in enumerate(cv2.split(resized_im)):
            # hist = cv2.calcHist(resized_im[:, :, channel], [0], None, [256], [0, 256])
            hist = cv2.calcHist([im_ch], [0], None, [256], [0, 256])
            # print(f"max diff {np.max(np.abs(hist-hist2))}")
            hist_all[channel, :] = hist[:, 0]

        hist_all = hist_all / np.max(hist_all)
        if start_hist: 
            end_hist = get_time()
            calc_hist_time += end_hist-start_hist
            gauss_start = get_time()
        else:
            gauss_start = None # Help syntax parser
        hist_all = cv2.GaussianBlur(hist_all, (7, 1), sigmaX=1.5, sigmaY=0.2)
        if gauss_start and h_start and resized_time:
            gauss_time = get_time() - gauss_start
            print(f"compute_histogram took {(get_time()-h_start)*1000:0.1f} msec. ", end="")
            print(f"from which calchist:{calc_hist_time*1000:0.1f}, "
              f"resizing:{resized_time*1000:0.1f}, "
              f"gauss:{gauss_time*1000:0.1f}")

        return hist_all

    def compute_histogram_Cpp(self, current_image, show_timings=False):
        # print(f"compute_histogram show_timings {show_timings}")
        h_start = get_time() if show_timings else None 
        # Compute steps based on input image resolution
        im_w, im_h = current_image.shape[1], current_image.shape[0]
        target_w = 800
        target_h = 600
        hist_x_step = max(1, int(im_w/target_w+0.5))
        hist_y_step = max(1, int(im_h/target_h+0.5))
        output_histogram = np.empty((3,256), dtype=np.uint32)
        qimview_cpp.compute_histogram(current_image, output_histogram, int(hist_x_step), int(hist_y_step))
        t1 = get_time() if show_timings else None
        hist_all = output_histogram.astype(np.float32)
        hist_all = hist_all / np.max(hist_all)
        hist_all = cv2.GaussianBlur(hist_all, (7, 1), sigmaX=1.5, sigmaY=0.2)
        if h_start and t1: 
            print(f"qimview_cpp.compute_histogram took {(get_time()-h_start)*1000:0.1f} ms, "
                                f"{(get_time()-t1)*1000:0.1f} ms")
        return hist_all

    def display_histogram(self, hist_all, id, painter, im_rect, show_timings=False):
        """
        :param painter:
        :param rect: displayed image area
        :return:
        """
        if hist_all is None:
            return
        histo_timings = show_timings
        #if histo_timings:
        h_start = get_time()
        # Histogram: keep constant width/height ratio
        display_ratio : float = 2.0
        # print(f'im_rect = {im_rect}')
        w, h = self.evt_width, self.evt_height
        width   : int = int( min(w/4*self._histo_scale, h/3*self._histo_scale))
        height  : int = int( width/display_ratio)
        start_x : int = w - width*id - 10
        start_y : int = h - 10
        margin  : int = 3

        rect_start = get_time() if histo_timings else None 
        rect = QtCore.QRect(start_x-margin, start_y-margin-height, width+2*margin, height+2*margin)
        self._histo_rect = rect
        # painter.fillRect(rect, QtGui.QBrush(QtGui.QColor(255, 255, 255, 128+64)))
        # Transparent light grey
        painter.fillRect(rect, QtGui.QColor(205, 205, 205, 128+32))
        rect_time = get_time()-rect_start if rect_start else None

        # print(f"current_image {current_image.shape} _image {self._image.shape}")
        # input_image = self._image
        path_time = 0

        pen = QtGui.QPen()
        pen.setWidth(2)

        qcolors = {
            0: QtGui.QColor(255, 50, 50, 255),
            1: QtGui.QColor(50, 255, 50, 255),
            2: QtGui.QColor(50, 50, 255, 255)
        }

        step_x = float(width) / 256
        step = 2
        x_range = np.array(range(0, 256, step))
        x_pos = start_x + x_range*step_x

        for channel in range(3):
            pen.setColor(qcolors[channel])
            painter.setPen(pen)
            # painter.setBrush(color)
            # print(f"histogram painting 1 took {get_time() - h_start} sec.")

            # print(f"histogram painting 2 took {get_time() - h_start} sec.")

            start_path = get_time() if histo_timings else None

            # apply a small Gaussian filtering to histogram curve
            path = QtGui.QPainterPath()

            y_pos = start_y - hist_all[channel, x_range]*height
            # polygon = QtGui.QPolygonF([QtCore.QPointF(x_pos[n], y_pos[n]) for n in range(len(x_range))])
            # path.addPolygon(polygon)
            path.moveTo(x_pos[0], y_pos[0])
            for n in range(1,len(x_range)):
                path.lineTo(x_pos[n], y_pos[n])
            painter.drawPath(path)
            path_time += get_time()-start_path if start_path else 0

        if rect_time: 
            print(f"display_histogram took {(get_time()-h_start)*1000:0.1f} msec. ", end='')
            print(f"from which path:{int(path_time*1000)}, rect:{int(rect_time*1000)}")


    def display_intensity_line(self, 
                               painter: QtGui.QPainter, 
                               im_rect: QtCore.QRect, 
                               line: np.ndarray,
                               channels : ImageFormat,
                                ) -> None:
        #if histo_timings:
        h_start = get_time()
        # print(f'im_rect = {im_rect}')
        w, h = self.evt_width, self.evt_height
        width    : int = im_rect.width()
        height   : int = int( h/5)
        start_x  : int = im_rect.x()
        margin_y : int = 2
        start_y  : int = h-margin_y

        rect = QtCore.QRect(start_x, start_y-height, width, height)
        painter.fillRect(rect, QtGui.QColor(205, 205, 205, 128+32))

        pen = QtGui.QPen()
        pen.setWidth(1)

        # Adapt for Bayer, Y, etc ...
        qcolors = {
            'R' : QtGui.QColor(240,  30,  30, 255),
            'G' : QtGui.QColor( 30, 240,  30, 255),
            'Gr': QtGui.QColor(130, 240,  30, 255),
            'Gb': QtGui.QColor( 30, 240, 130, 255),
            'B' : QtGui.QColor( 30,  30, 240, 255),
            'Y' : QtGui.QColor( 30,  30,  30, 255),
        }
        colors = {
            ImageFormat.CH_RGB  : ['R','G','B'],
            ImageFormat.CH_BGR  : ['B','G','R'],
            ImageFormat.CH_RGGB : ['R','Gr','Gb','B'],
            ImageFormat.CH_GRBG : ['Gr','R','B','Gb'],
            ImageFormat.CH_GBRG : ['Gb','B','R','Gr'],
            ImageFormat.CH_BGGR : ['B','Gb','Gr','R'],
        }[channels]
        assert line.shape[1] == len(colors), f"Error: Mismatch between imageformat and number of channels"

        nb_values = line.shape[0]
        step_x = float(width) / nb_values
        x_range = np.array(range(0, nb_values))
        x_pos = start_x + (x_range+0.5)*step_x

        max_val = np.max(line)
        line = line.astype(np.float32)
        in_margin = 2
        in_start_y = start_y - in_margin
        in_height  = height - 2*in_margin
        for channel in range(len(colors)):
            pen.setColor(qcolors[colors[channel]])
            painter.setPen(pen)
            # apply a small Gaussian filtering to histogram curve
            path = QtGui.QPainterPath()
            y_pos = (in_start_y - line[:,channel]*(in_height/max_val)+0.5).astype(np.uint32)
            path.moveTo(x_pos[0], y_pos[0])
            for n in range(1,len(x_range)):
                path.lineTo(x_pos[n], y_pos[n])
            painter.drawPath(path)





#
#
# started from https://cyrille.rossant.net/2d-graphics-rendering-tutorial-with-pyopengl/
#
# check also https://doc.qt.io/archives/4.6/opengl-overpainting.html
#

from qimview.utils.qt_imports import *
from qimview.utils.viewer_image import *
from qimview.utils.utils import clip_value
from qimview.utils.utils import get_time
from qimview.tests_utils.qtdump import *
# Renaming manually since syntax checker has issues with cv2
import cv2
# from cv2 import resize          as opencv_resize
# from cv2 import add             as opencv_add
# from cv2 import subtract        as opencv_subtract
# from cv2 import addWeighted     as opencv_addWeighted
# from cv2 import convertScaleAbs as opencv_convertScaleAbs
# from cv2 import INTER_NEAREST   as opencv_INTER_NEAREST
# from cv2 import INTER_AREA      as opencv_INTER_AREA
import numpy as np
from typing import Tuple, Optional

try:
    import qimview_cpp
except Exception as e:
    HAS_CPPBIND = False
    print("Failed to load qimview_cpp: {}".format(e))
else:
    HAS_CPPBIND = True
print("Do we have cpp binding ? {}".format(HAS_CPPBIND))

from qimview.image_viewers import ImageFilterParameters
from .image_viewer import (ImageViewer, trace_method,)


# the opengl version is a bit slow for the moment, due to the texture generation
# BaseWidget = QOpenGLWidget 
BaseWidget = QtWidgets.QWidget

class QTImageViewer(ImageViewer, BaseWidget):

    def __init__(self, parent : Optional[QtWidgets.QWidget] = None, event_recorder=None):
        BaseWidget.__init__(self, parent)
        ImageViewer.__init__(self, self)
        self.event_recorder = event_recorder
        self.setMouseTracking(True)
        self.anti_aliasing = True
        size_policy = QtWidgets.QSizePolicy()
        size_policy.setHorizontalPolicy(QtWidgets.QSizePolicy.Policy.Ignored)
        size_policy.setVerticalPolicy  (QtWidgets.QSizePolicy.Policy.Ignored)
        self.setSizePolicy(size_policy)
        # self.setAlignment(QtCore.Qt.AlignCenter )
        self.output_crop = np.array([0., 0., 1., 1.], dtype=np.float32)
        self.zoom_center = np.array([0.5, 0.5, 0.5, 0.5])
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        self.paint_cache      = None
        self.paint_diff_cache = None
        self.diff_image       = None

        # self.display_timing = False
        if BaseWidget is QOpenGLWidget:
            self.setAutoFillBackground(True)

        # TODO: how can I set the background color to black without impacting display speed?
        # p = self.palette()
        # p.setColor(self.backgroundRole(), QtCore.Qt.black) 
        # self.setPalette(p)
        # self.setAutoFillBackground(True)

        self.verbose = False
        # self.trace_calls = True

    def apply_zoom(self, crop):
        if self._image is None: return crop
        (height, width) = self._image.data.shape[:2]
        # print(f"height, width = {height, width}")
        # Apply zoom
        coeff = 1.0/self.new_scale(-self.mouse_zoom_displ.y(), height)
        # zoom from the center of the image
        center = self.zoom_center
        new_crop = center + (crop - center) * coeff

        # print("new crop zoom 1 {}".format(new_crop))

        # allow crop increase based on the available space
        label_width = self.width()
        # print(f"label_width {label_width}")
        label_height = self.height()

        new_width = width * coeff
        new_height = height * coeff

        ratio_width = float(label_width) / new_width
        ratio_height = float(label_height) / new_height

        # print(f" ratio_width {ratio_width} ratio_height {ratio_height}")
        ratio = min(ratio_width, ratio_height)

        if ratio_width<ratio_height:
            # margin to increase height
            margin_pixels = label_height/ratio - new_height
            margin_height = margin_pixels/height
            new_crop[1] -= margin_height/2
            new_crop[3] += margin_height/2
        else:
            # margin to increase width
            margin_pixels = label_width/ratio - new_width
            margin_width = margin_pixels/width
            new_crop[0] -= margin_width/2
            new_crop[2] += margin_width/2
        # print("new crop zoom 2 {}".format(new_crop))

        return new_crop

    def apply_translation(self, crop):
        """
        :param crop:
        :return: the new crop
        """
        # Apply translation
        diff_x, diff_y = self.new_translation()
        diff_y = - diff_y
        # print(" new translation {} {}".format(diff_x, diff_y))
        # apply the maximal allowed translation
        tr_x = float(diff_x) / self.width()
        tr_y = float(diff_y) / self.height()
        tr_x = clip_value(tr_x, crop[2]-1, crop[0])
        tr_y = clip_value(tr_y, crop[3]-1, crop[1])
        # normalized position relative to the full image
        crop[0] -= tr_x
        crop[1] -= tr_y
        crop[2] -= tr_x
        crop[3] -= tr_y

    def check_translation(self):
        """
        This method computes the translation really applied based on the current requested translation
        :return:
        """
        # Apply zoom
        crop = self.apply_zoom(self.output_crop)

        # Compute the translation that is really applied after applying the constraints
        diff_x, diff_y = self.new_translation()
        diff_y = - diff_y
        # print(" new translation {} {}".format(diff_x, diff_y))
        # apply the maximal allowed translation
        w, h = self.width(), self.height()
        diff_x = clip_value(diff_x, w*(crop[2]-1), w*(crop[0]))
        diff_y = - clip_value(diff_y, h*(crop[3]-1), h*(crop[1]))
        # normalized position relative to the full image
        return diff_x, diff_y

    def update_crop(self):
        # Apply zoom
        new_crop = self.apply_zoom(self.output_crop)
        # print(f"update_crop {self.output_crop} --> {new_crop}")
        # Apply translation
        self.apply_translation(new_crop)
        new_crop = np.clip(new_crop, 0, 1)
        # print("move new crop {}".format(new_crop))
        # print(f"output_crop {self.output_crop} new crop {new_crop}")
        return new_crop

    def update_crop_new(self):
        # 1. transform crop to display coordinates
        
        # Apply zoom
        new_crop = self.apply_zoom(self.output_crop)
        # print(f"update_crop {self.output_crop} --> {new_crop}")
        # Apply translation
        self.apply_translation(new_crop)
        new_crop = np.clip(new_crop, 0, 1)
        # print("move new crop {}".format(new_crop))
        # print(f"output_crop {self.output_crop} new crop {new_crop}")
        return new_crop

    def apply_filters(self, current_image: ViewerImage) -> np.ndarray:
        self.print_log(f"current_image.data.shape {current_image.data.shape}")
        # return current_image

        self.start_timing(title='apply_filters()')

        # Output RGB from input
        if self._image is None: return current_image
        ch = self._image.channels
        if HAS_CPPBIND:
            channels = current_image.channels
            black_level = self.filter_params.black_level.float
            white_level = self.filter_params.white_level.float
            g_r_coeff = self.filter_params.g_r.float
            g_b_coeff = self.filter_params.g_b.float
            saturation = self.filter_params.saturation.float
            max_value = ((1<<current_image.precision)-1)
            max_type = 1  # not used
            gamma = self.filter_params.gamma.float  # not used

            rgb_image = np.empty((current_image.data.shape[0], current_image.data.shape[1], 3), dtype=np.uint8)
            time1 = get_time()
            ok = False
            if ch in ImageFormat.CH_RAWFORMATS() or ch in ImageFormat.CH_RGBFORMATS():
                cases = {
                    'uint8':  { 'func': qimview_cpp.apply_filters_u8_u8  , 'name': 'apply_filters_u8_u8'},
                    'uint16': { 'func': qimview_cpp.apply_filters_u16_u8, 'name': 'apply_filters_u16_u8'},
                    'uint32': { 'func': qimview_cpp.apply_filters_u32_u8, 'name': 'apply_filters_u32_u8'},
                    'int16': { 'func': qimview_cpp.apply_filters_s16_u8, 'name': 'apply_filters_s16_u8'},
                    'int32': { 'func': qimview_cpp.apply_filters_s32_u8, 'name': 'apply_filters_s32_u8'}
                }
                if current_image.data.dtype.name in cases:
                    func = cases[current_image.data.dtype.name]['func']
                    name = cases[current_image.data.dtype.name]['name']
                    self.print_log(f"qimview_cpp.{name}(current_image, rgb_image, channels, "
                          f"black_level={black_level}, white_level={white_level}, "
                          f"g_r_coeff={g_r_coeff}, g_b_coeff={g_b_coeff}, "
                          f"max_value={max_value}, max_type={max_type}, gamma={gamma})")
                    ok = func(current_image.data, rgb_image, channels, black_level, white_level, g_r_coeff,
                                g_b_coeff, max_value, max_type, gamma, saturation)
                    self.add_time(f'{name}()',time1, force=True, title='apply_filters()')
                else:
                    print(f"apply_filters() not available for {current_image.data.dtype} data type !")
            else:
                cases = {
                    'uint8': { 'func': qimview_cpp.apply_filters_scalar_u8_u8, 'name': 'apply_filters_scalar_u8_u8'},
                    'uint16': { 'func': qimview_cpp.apply_filters_scalar_u16_u8, 'name': 'apply_filters_scalar_u16_u8'},
                    'int16': { 'func': qimview_cpp.apply_filters_scalar_s16_u8, 'name': 'apply_filters_scalar_s16_u8'},
                    'uint32': { 'func': qimview_cpp.apply_filters_scalar_u32_u8, 'name': 'apply_filters_scalar_u32_u8'},
                    'float64': { 'func': qimview_cpp.apply_filters_scalar_f64_u8, 'name': 'apply_filters_scalar_f64_u8'},
                }
                if current_image.data.dtype.name.startswith('float'):
                    max_value = 1.0
                if current_image.data.dtype.name in cases:
                    func = cases[current_image.data.dtype.name]['func']
                    name = cases[current_image.data.dtype.name]['name']
                    self.print_log(f"qimview_cpp.{name}(current_image, rgb_image, "
                          f"black_level={black_level}, white_level={white_level}, "
                          f"max_value={max_value}, max_type={max_type}, gamma={gamma})")
                    ok = func(current_image.data, rgb_image, black_level, white_level, max_value, max_type, gamma)
                    self.add_time(f'{name}()', time1, force=True, title='apply_filters()')
                else:
                    print(f"apply_filters_scalar() not available for {current_image.data.dtype} data type !")
            if not ok:
                self.print_log("Failed running wrap_num.apply_filters_u16_u8 ...", force=True)
        else:
            # self.print_log("current channels {}".format(ch))
            if ch in ImageFormat.CH_RAWFORMATS():
                channel_pos = channel_position[current_image.channels]
                self.print_log("Converting to RGB")
                # convert Bayer to RGB
                rgb_image = np.empty((current_image.data.shape[0], current_image.data.shape[1], 3), 
                                        dtype=current_image.data.dtype)
                rgb_image[:, :, 0] = current_image.data[:, :, channel_pos['r']]
                rgb_image[:, :, 1] = (current_image.data[:, :, channel_pos['gr']]+current_image.data[:, :, channel_pos['gb']])/2
                rgb_image[:, :, 2] = current_image.data[:, :, channel_pos['b']]
            else:
                if ch == ImageFormat.CH_Y:
                    # Transform to RGB is it a good idea?
                    rgb_image = np.empty((current_image.data.shape[0], current_image.data.shape[1], 3), 
                                            dtype=current_image.data.dtype)
                    rgb_image[:, :, 0] = current_image.data
                    rgb_image[:, :, 1] = current_image.data
                    rgb_image[:, :, 2] = current_image.data
                else:
                    rgb_image = current_image.data

            # Use cv2.convertScaleAbs(I,a,b) function for fast processing
            # res = sat(|I*a+b|)
            # if current_image is not in 8 bits, we need to rescale
            min_val = self.filter_params.black_level.float
            max_val = self.filter_params.white_level.float

            if min_val != 0 or max_val != 1 or current_image.precision!=8:
                min_val = self.filter_params.black_level.float
                max_val = self.filter_params.white_level.float
                # adjust levels to precision
                precision = current_image.precision
                min_val = min_val*((1 << precision)-1)
                max_val = max_val*((1 << precision)-1)
                if rgb_image.dtype == np.uint32:
                    # Formula a bit complicated, we need to be careful with unsigned processing
                    rgb_image =np.clip(((np.clip(rgb_image, min_val, None) - min_val)*(255/(max_val-min_val)))+0.5,
                                       None, 255).astype(np.uint8)
                else:
                    # to rescale: add min_val and multiply by (max_val-min_val)/255
                    if min_val != 0:
                        rgb_image = cv2.add(rgb_image, (-min_val, -min_val, -min_val, 0))
                    rgb_image = cv2.convertScaleAbs(rgb_image, alpha=255. / float(max_val - min_val), beta=0)

            # # if gamma changed
            # if self.filter_params.gamma.value != self.filter_params.gamma.default_value and work_image.dtype == np.uint8:
            #     gamma_coeff = self.filter_params.gamma.float
            #     # self.gamma_label.setText("Gamma  {}".format(gamma_coeff))
            #     invGamma = 1.0 / gamma_coeff
            #     table = np.array([((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]).astype("uint8")
            #     work_image = cv2.LUT(work_image, table)

        self.print_timing(title='apply_filters()')
        return rgb_image

    def viewer_update(self):
        if BaseWidget is QOpenGLWidget:
            self.paint_image()
            self.repaint()
        else:
            self.update()

    def draw_overlay_separation(self, cropped_image_shape, rect : QtCore.QRect, painter):
        (height, width) = cropped_image_shape[:2]
        im_x = int((self.mouse_pos.x() - rect.x())/rect.width()*width)
        im_x = max(0, min(width - 1, im_x))
        # Set position at the beginning of the pixel
        pos_from_im_x = int(im_x*rect.width()/width + rect.x())
        pen_width = 2
        color = QtGui.QColor(255, 255, 0 , 128)
        pen = QtGui.QPen()
        pen.setColor(color)
        pen.setWidth(pen_width)
        painter.setPen(pen)
        painter.drawLine(pos_from_im_x, rect.y(), pos_from_im_x, rect.y() + rect.height())

    def draw_cursor(self, 
                    cropped_image_shape, 
                    crop_xmin, 
                    crop_ymin, 
                    rect : QtCore.QRect, 
                    painter, 
                    full=False
        ) -> Optional[Tuple[int, int]]:
        """
        :param cropped_image_shape: dimensions of current crop
        :param crop_xmin: left pixel of current crop
        :param crop_ymin: top pixel of current crop
        :param rect: displayed image area
        :param painter:
        :return:
            tuple: (posx, posy) image pixel position of the cursor, if None cursor is out of image
        """
        # Draw cursor
        if self.display_timing: self.start_timing()
        # get image position
        (height, width) = cropped_image_shape[:2]
        im_x = int((self.mouse_pos.x() -rect.x())/rect.width()*width)
        im_y = int((self.mouse_pos.y() -rect.y())/rect.height()*height)

        pos_from_im_x = int((im_x+0.5)*rect.width()/width +rect.x())
        pos_from_im_y = int((im_y+0.5)*rect.height()/height+rect.y())

        # ratio = self.screen().devicePixelRatio()
        # print("ratio = {}".format(ratio))
        pos_x = pos_from_im_x  # *ratio
        pos_y = pos_from_im_y  # *ratio
        length_percent = 0.04
        # use percentage of the displayed image dimensions
        length = int(max(self.width(),self.height())*length_percent)
        pen_width = 2 if full else 3
        color = QtGui.QColor(0, 255, 255, 200)
        pen = QtGui.QPen()
        pen.setColor(color)
        pen.setWidth(pen_width)
        painter.setPen(pen)
        if not full:
            painter.drawLine(pos_x-length, pos_y, pos_x+length, pos_y)
            painter.drawLine(pos_x, pos_y-length, pos_x, pos_y+length)
        else:
            painter.drawLine(rect.x(), pos_y, rect.x()+rect.width(), pos_y)
            painter.drawLine(pos_x, rect.y(), pos_x, rect.y()+rect.height())

        # Update text
        if im_x>=0 and im_x<cropped_image_shape[1] and im_y>=0 and im_y<cropped_image_shape[0]:
            # values = cropped_image[im_y, im_x]
            im_x += crop_xmin
            im_y += crop_ymin
            im_pos = (im_x, im_y)
        else:
            im_pos = None
        if self.display_timing: self.print_timing()
        return im_pos

    def get_difference_image(self, verbose=True) -> Optional[ViewerImage]:

        factor = self.filter_params.imdiff_factor.float
        if self.paint_diff_cache is not None:
            use_cache = self.paint_diff_cache['imid'] == self.image_id and \
                        self.paint_diff_cache['imrefid'] == self.image_ref_id and \
                        self.paint_diff_cache['factor'] == factor
        else:
            use_cache = False

        if self._image is None or self._image_ref is None: return None
        if not use_cache:
            im1 = self._image.data
            im2 = self._image_ref.data
            # TODO: get factor from parameters ...
            # factor = int(self.diff_color_slider.value())
            print(f'factor = {factor}')
            print(f' im1.dtype {im1.dtype} im2.dtype {im2.dtype}')
            # Fast OpenCV code
            start = get_time()
            # positive diffs in unsigned 8 bits, OpenCV puts negative values to 0
            try:
                if im1.dtype.name == 'uint8' and im2.dtype.name == 'uint8':
                    diff_plus = cv2.subtract(im1, im2)
                    diff_minus = cv2.subtract(im2, im1)
                    res = cv2.addWeighted(diff_plus, factor, diff_minus, -factor, 127)
                    if verbose:
                        print(f" qtImageViewer.difference_image()  took {int((get_time() - start)*1000)} ms")
                        vmin = np.min(res)
                        vmax = np.max(res)
                        print(f"min-max diff = {vmin} - {vmax}")
                        histo,_ = np.histogram(res, bins=int(vmax-vmin+0.5), range=(vmin, vmax))
                        sum = 0
                        for v in range(vmin,vmax):
                            if v!=127:
                                nb = histo[v-vmin]
                                if nb >0:
                                    print(f"{v-127}:{nb} ",end='')
                                    sum += nb
                        print('')
                        print(f'nb pixel diff  {sum}')
                    res = ViewerImage(res,  precision=self._image.precision, 
                                            downscale=self._image.downscale,
                                            channels=self._image.channels)
                    self.paint_diff_cache = {  'imid': self.image_id, 'imrefid': self.image_ref_id, 
                        'factor': self.filter_params.imdiff_factor.float
                    }
                    self.diff_image = res
                else:
                    d = (im1.astype(np.float32)-im2.astype(np.float32))*factor
                    d[d<-127] = -127
                    d[d>128] = 128
                    d = (d+127).astype(np.uint8)*255
                    res = ViewerImage(d,  precision=8, 
                                            downscale=self._image.downscale,
                                            channels=self._image.channels)
                    self.paint_diff_cache = {  'imid': self.image_id, 'imrefid': self.image_ref_id, 
                        'factor': self.filter_params.imdiff_factor.float
                    }
                    self.diff_image = res
            except Exception as e:
                print(f"Error {e}")
                res = (im1!=im2).astype(np.uint8)*255
                res = ViewerImage(res,  precision=8, 
                                        downscale=self._image.downscale,
                                        channels=ImageFormat.CH_Y)
                self.diff_image = res

        return self.diff_image

    def paint_image(self):
        # print(f"paint_image display_timing {self.display_timing}")
        if self.trace_calls: t = trace_method(self.tab)
        self.start_timing()
        time0 = time1 = get_time()

        label_width = self.size().width()
        label_height = self.size().height()

        show_diff = self.show_image_differences and self._image is not self._image_ref and \
                    self._image is not None and \
                    self._image_ref is not None and self._image.data.shape == self._image_ref.data.shape

        c = self.update_crop()
        # check paint_cache
        if self.paint_cache is not None:
            use_cache = self.paint_cache['imid'] == self.image_id and \
                        np.array_equal(self.paint_cache['crop'],c) and \
                        self.paint_cache['labelw'] == label_width and \
                        self.paint_cache['labelh'] == label_height and \
                        self.paint_cache['filterp'].is_equal(self.filter_params) and \
                        (self.paint_cache['showhist'] == self.show_histogram or not self.show_histogram) and \
                        self.paint_cache['show_diff'] == show_diff and \
                        self.paint_cache['antialiasing'] == self.antialiasing and \
                        not self.show_overlay
        else:
            use_cache = False

        # if show_diff, compute the image difference (put it in cache??)
        if show_diff:
            # Cache does not work well with differences
            use_cache = False
            # don't save the difference
            current_image = self.get_difference_image()
        else:
            current_image = self._image

        if current_image is None: return
        precision  = current_image.precision
        downscale  = current_image.downscale
        channels   = current_image.channels

        # TODO: get data based on the display ratio?
        image_data = current_image.data

        # could_use_cache = use_cache
        # if could_use_cache:
        #     print(" Could use cache here ... !!!")
        # use_cache = False

        do_crop = (c[2] - c[0] != 1) or (c[3] - c[1] != 1)
        h, w  = image_data.shape[:2]
        if do_crop:
            crop_xmin = int(np.round(c[0] * w))
            crop_xmax = int(np.round(c[2] * w))
            crop_ymin = int(np.round(c[1] * h))
            crop_ymax = int(np.round(c[3] * h))
            image_data = image_data[crop_ymin:crop_ymax, crop_xmin:crop_xmax]
        else:
            crop_xmin = crop_ymin = 0
            crop_xmax = w
            crop_ymax = h

        cropped_image_shape = image_data.shape
        self.add_time('crop', time1)

        # time1 = get_time()
        image_height, image_width  = image_data.shape[:2]
        ratio_width = float(label_width) / image_width
        ratio_height = float(label_height) / image_height
        ratio = min(ratio_width, ratio_height)
        display_width = int(round(image_width * ratio))
        display_height = int(round(image_height * ratio))

        if self.show_overlay and self._image_ref is not self._image and self._image_ref and self._image and \
            self._image.data.shape == self._image_ref.data.shape:
            # to create the overlay rapidly, we will mix the two images based on the current cursor position
            # 1. convert cursor position to image position
            (height, width) = cropped_image_shape[:2]
            # compute rect
            rect = QtCore.QRect(0, 0, display_width, display_height)
            devRect = QtCore.QRect(0, 0, self.evt_width, self.evt_height)
            rect.moveCenter(devRect.center())
            im_x = int((self.mouse_pos.x() - rect.x()) / rect.width() * width)
            im_x = max(0,min(width-1, im_x))
            # We need to have a copy here .. slow, better option???
            image_data = np.copy(image_data)
            image_data[:, :im_x] = self._image_ref.data[crop_ymin:crop_ymax, crop_xmin:(crop_xmin+im_x)]

        resize_applied = False
        if not use_cache:
            anti_aliasing = ratio < 1
            #self.print_log("ratio is {:0.2f}".format(ratio))
            use_opencv_resize : bool = anti_aliasing
            # enable this as optional?
            # opencv_downscale_interpolation = cv2.fast_interpolation
            cv2.fast_interpolation = cv2.INTER_NEAREST
            if self.antialiasing:
                opencv_downscale_interpolation = cv2.INTER_AREA
            else:
                opencv_downscale_interpolation = cv2.INTER_NEAREST
            # opencv_upscale_interpolation   = cv2.INTER_LINEAR
            opencv_upscale_interpolation   = cv2.fast_interpolation
            # self.print_time('several settings', time1, start_time)

            # self.print_log("use_opencv_resize {} channels {}".format(use_opencv_resize, current_image.channels))
            # if ratio<1 we want anti aliasing and we want to resize as soon as possible to reduce computation time
            if use_opencv_resize and not resize_applied and channels == ImageFormat.CH_RGB:

                prev_shape = image_data.shape
                initial_type = image_data.dtype
                if image_data.dtype != np.uint8:
                    print(f"image_data type {type(image_data)} {image_data.dtype}")
                    image_data = image_data.astype(np.float32)

                # if ratio is >2, start with integer downsize which is much faster
                # we could add this condition opencv_downscale_interpolation==cv2.INTER_AREA
                if ratio<=0.5:
                    if image_data.shape[0]%2!=0 or image_data.shape[1]%2 !=0:
                        # clip image to multiple of 2 dimension
                        image_data = image_data[:2*(image_data.shape[0]//2),:2*(image_data.shape[1]//2)]
                    start_0 = get_time()
                    resized_image = cv2.resize(image_data, (image_width>>1, image_height>>1),
                                            interpolation=opencv_downscale_interpolation)
                    if self.display_timing:
                        print(f' === qtImageViewer: ratio {ratio:0.2f} paint_image() OpenCV resize from '
                            f'{current_image.data.shape} to '
                            f'{resized_image.shape} --> {int((get_time()-start_0)*1000)} ms')
                    image_data = resized_image
                    if ratio<=0.25:
                        if image_data.shape[0]%2!=0 or image_data.shape[1]%2 !=0:
                            # clip image to multiple of 2 dimension
                            image_data = image_data[:2*(image_data.shape[0]//2),:2*(image_data.shape[1]//2)]
                        start_0 = get_time()
                        resized_image = cv2.resize(image_data, (image_width>>2, image_height>>2),
                                                interpolation=opencv_downscale_interpolation)
                        if self.display_timing:
                            print(f' === qtImageViewer: ratio {ratio:0.2f} paint_image() OpenCV resize from '
                                f'{current_image.data.shape} to '
                                f'{resized_image.shape} --> {int((get_time()-start_0)*1000)} ms')
                        image_data = resized_image

                time1 = get_time()
                start_0 = get_time()
                resized_image = cv2.resize(image_data, (display_width, display_height),
                                        interpolation=opencv_downscale_interpolation)
                if self.display_timing:
                    print(f' === qtImageViewer: paint_image() OpenCV resize from {image_data.shape} to '
                        f'{resized_image.shape} --> {int((get_time()-start_0)*1000)} ms')

                image_data = resized_image.astype(initial_type)
                resize_applied = True
                self.add_time('cv2.resize',time1)

            current_image = ViewerImage(image_data,  precision=precision, downscale=downscale, channels=channels)
            if self.show_stats and self._image:
                # Output RGB from input
                ch = self._image.channels
                data_shape = current_image.data.shape
                if len(data_shape)==2:
                    print(f"input average {np.average(current_image.data)}")
                if len(data_shape)==3:
                    for c in range(data_shape[2]):
                        print(f"input average ch {c} {np.average(current_image.data[:,:,c])}")
            current_image = self.apply_filters(current_image)

            # Compute the histogram here, with the smallest image!!!
            if self.show_histogram:
                # previous version only python with its modules
                # histograms  = self.compute_histogram    (current_image, show_timings=self.display_timing)
                # new version with bound C++ code and openMP: much faster
                histograms = self.compute_histogram_Cpp(current_image, show_timings=self.display_timing)
            else:
                histograms = None

            # try to resize anyway with opencv since qt resizing seems too slow
            if not resize_applied and BaseWidget is not QOpenGLWidget:
                time1 = get_time()
                start_0 = get_time()
                prev_shape = current_image.shape
                current_image = cv2.resize(current_image, (display_width, display_height),
                                           interpolation=opencv_upscale_interpolation)
                if self.display_timing:
                    print(f' === qtImageViewer: paint_image() OpenCV resize from {prev_shape} to '
                        f'{(display_height, display_width)} --> {int((get_time()-start_0)*1000)} ms')
                    self.add_time('cv2.resize',time1)

            # no need for more resizing
            resize_applied = True

            # Conversion from numpy array to QImage
            # version 1: goes through PIL image
            # version 2: use QImage constructor directly, faster
            # time1 = get_time()

        else:
            resize_applied = True
            if self.paint_cache:
                current_image = self.paint_cache['current_image']
                histograms = self.paint_cache['histograms']
            # histograms2 = self.paint_cache['histograms2']

        # if could_use_cache:
        #     print(f" ======= current_image equal ? {np.array_equal(self.paint_cache['current_image'],current_image)}")

        if not use_cache and not self.show_overlay:
            # cache_time = get_time()
            fp = ImageFilterParameters()
            fp.copy_from(self.filter_params)
            self.paint_cache = {
                'imid': self.image_id,
                'imrefid': self.image_ref_id,
                'crop': c, 'labelw': label_width, 'labelh': label_height,
                'filterp': fp, 'showhist': self.show_histogram,
                'histograms': histograms, 
                # 'histograms2': histograms2, 
                'current_image': current_image,
                'show_diff' : show_diff,
                'antialiasing': self.antialiasing
                }
            # print(f"create cache data took {int((get_time() - cache_time) * 1000)} ms")

        if not current_image.flags['C_CONTIGUOUS']:
            current_image = np.require(current_image, np.uint8, 'C')
        qimage = QtGui.QImage(current_image.data, current_image.shape[1], current_image.shape[0],
                                    current_image.strides[0], QtGui.QImage.Format_RGB888)
        # self.add_time('QtGui.QPixmap',time1)

        assert resize_applied, "Image resized should be applied at this point"

        if self._save_image_clipboard and self._clipboard:
            self.print_log("exporting to clipboard")
            self._clipboard.setImage(qimage, mode=QtGui.QClipboard.Mode.Clipboard)

        painter : QtGui.QPainter = QtGui.QPainter()

        painter.begin(self)
        if BaseWidget is QOpenGLWidget:
            painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)

        # TODO: check that this condition is not needed
        if BaseWidget is QOpenGLWidget:
            rect : QtCore.QRect = QtCore.QRect(0,0, display_width, display_height)
        else:
            rect : QtCore.QRect = qimage.rect()
        devRect = QtCore.QRect(0, 0, self.evt_width, self.evt_height)
        rect.moveCenter(devRect.center())

        time1 = get_time()
        if BaseWidget is QOpenGLWidget:
            painter.drawImage(rect, qimage)
        else:
            painter.drawImage(rect.topLeft(), qimage)
        self.add_time('painter.drawImage',time1)

        if self.show_overlay:
            self.draw_overlay_separation(cropped_image_shape, rect, painter)

        # Draw cursor
        im_pos = None
        if self.show_cursor:
            im_pos = self.draw_cursor(cropped_image_shape, 
                                      crop_xmin, 
                                      crop_ymin, 
                                      rect, 
                                      painter, 
                                      full = self.show_intensity_line,
                                      )

        if self.show_intensity_line and self._image:
            (height, width) = cropped_image_shape[:2]
            im_y = int((self.mouse_pos.y() -rect.y())/rect.height()*height)
            im_y += crop_ymin
            im_shape = self._image.data.shape
            # Horizontal display
            if im_y>=0 and im_y<im_shape[0] and crop_xmin>=0 and crop_xmin+cropped_image_shape[1]<=im_shape[1]:
                line = self._image.data[im_y, crop_xmin:crop_xmin+cropped_image_shape[1]]
                self.display_intensity_line(
                    painter, 
                    rect, 
                    line,
                    channels = self._image.channels,
                    )

        self.display_text(painter, self.display_message(im_pos, ratio*self.devicePixelRatio()))

        # draw histogram
        if self.show_histogram:
            self.display_histogram(histograms, 1,  painter, rect, show_timings=self.display_timing)
            # self.display_histogram(histograms2, 2, painter, rect, show_timings=self.display_timing)

        painter.end()
        self.print_timing()

        if self.display_timing:
            print(f" paint_image took {int((get_time()-time0)*1000)} ms")

    def show(self):
        if BaseWidget==QOpenGLWidget:
            self.update()
        BaseWidget.show(self)

    def paintEvent(self, event):
        # print(f" qtImageViewer.paintEvent() {self.image_name}")
        if self.trace_calls:
            t = trace_method(self.tab)
        # try:
        if self._image is not None:
            self.paint_image()
        # except Exception as e:
        #     print(f"Failed paint_image() {e}")

    def resizeEvent(self, event):
        """Called upon window resizing: reinitialize the viewport.
        """
        if self.trace_calls:
            t = trace_method(self.tab)
        self.print_log(f"resize {event.size()}  self {self.width()} {self.height()}")
        self.evt_width = event.size().width()
        self.evt_height = event.size().height()
        BaseWidget.resizeEvent(self, event)
        self.print_log(f"resize {event.size()}  self {self.width()} {self.height()}")

    def mousePressEvent  (self, event:  QtGui.QMouseEvent) -> None: 
        self._mouse_events.mouse_press_event(event)
    def mouseReleaseEvent(self, event : QtGui.QMouseEvent) -> None: 
        self._mouse_events.mouse_release_event(event)
    def mouseMoveEvent   (self, event : QtGui.QMoveEvent)  -> None: 
        self._mouse_events.mouse_move_event(event)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) ->None:
        self._mouse_events.mouse_double_click_event(event)

    def wheelEvent(self, event):
        self._mouse_events.mouse_wheel_event(event)

    def event(self, evt):
        if self.event_recorder is not None:
            self.event_recorder.store_event(self, evt)
        return BaseWidget.event(self, evt)

    def keyPressEvent(self, event):
        self.key_press_event(event, wsize=self.size())

    def keyReleaseEvent(self, evt):
        self.print_log(f"evt {evt.type()}")



#
# use %load file in ipython to exexute this file
#
import qimview_cpp
import numpy as np

# if False:
# 	a = (np.random.rand(2000,2000,3)*100).astype(np.float64)
# 	b = a.copy()
# 	%time b += 1
# 	%time qimview_cpp.increment_3d_omp(b)
# 	%time qimview_cpp.increment_3d(b)


a = (np.random.rand(2000,2000,4)*4095).astype(np.uint16)

# try the function
channels = 4 #ImageFormat.CH_RGGB
white_level = 1
black_level = 0
g_r_coeff = 1.5
g_b_coeff = 1
max_value = 4095
max_type = 1 # not used
gamma = 1 # not used

b = np.copy(a)
qimview_cpp.apply_filters_u16_u8(b, channels, black_level, white_level, g_r_coeff, g_b_coeff, max_value, max_type, gamma)


print(a[0,0,:])
print(a[0,0,:].astype(np.float64)/max_value*255)

print(b[0,0,:])




# from https://numpy.org/doc/stable/user/basics.subclassing.html

from __future__ import annotations
import numpy as np
import cv2
from .utils import get_time
from enum import Enum, IntEnum
from typing import Tuple

class ImageFormat(IntEnum):
    CH_RGB  = 1 
    """RGB colour image"""
    CH_BGR  = 2 
    "RBG with inverted order"
    CH_Y    = 3 
    "One channel greyscale image"
    CH_RGGB = 4 
    "phase 0, bayer 2"
    CH_GRBG = 5 
    "phase 1, bayer 3" 
    CH_GBRG = 6 
    "phase 2, bayer 0"
    CH_BGGR = 7 
    "phase 3, bayer 1"

    @staticmethod
    def CH_RAWFORMATS() -> Tuple[ImageFormat, ...]:
        return (ImageFormat.CH_RGGB, ImageFormat.CH_GRBG, ImageFormat.CH_GBRG, ImageFormat.CH_BGGR,)
    
    @staticmethod
    def CH_RGBFORMATS() -> Tuple[ImageFormat, ...]:
        return (ImageFormat.CH_RGB, ImageFormat.CH_BGR,)

    @staticmethod
    def CH_SCALARFORMATS() -> Tuple[ImageFormat, ...]:
        return (ImageFormat.CH_Y,)

channel_position = {
    ImageFormat.CH_RGGB: {'r' :0, 'gr':1, 'gb':2, 'b' :3},
    ImageFormat.CH_GRBG: {'gr':0, 'r' :1, 'b' :2, 'gb':3},
    ImageFormat.CH_GBRG: {'gb':0, 'b' :1, 'r' :2, 'gr':3},
    ImageFormat.CH_BGGR: {'b' :0, 'gb':1, 'gr':2, 'r' :3},
}

class ViewerImage:
    """
    Own image class that inherits from np.ndarray
    """

    def __init__(self, 
                 input_array : np.ndarray, 
                 precision : int =8, 
                 downscale=1, 
                 channels: ImageFormat = ImageFormat.CH_RGB,
                 ):
        """
        :param input_array:
        :param precision: Integer image precision (number of bits)
        """
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        self._data = input_array
        # add the new attribute to the created instance
        self.precision : int         = precision
        self.downscale : int         = downscale
        self.channels  : ImageFormat = channels
        self.filename  : str         = None
        self.data_reduced_2 = None
        self.data_reduced_4 = None


    @property
    def data(self) -> np.ndarray :
        return self._data

    @data.setter
    def data(self, v : np.ndarray):
        self._data = v

    def reduce_half(self, input_data, interpolation=cv2.INTER_AREA, display_timing=False):
        image_height, image_width  = input_data.shape[:2]
        start_0 = get_time()
        if image_height % 2 != 0 or image_width % 2 != 0:
            # clip image to multiple of 2 dimension
            input_2 = input_data[:2*(image_height//2),:2*(image_width[1]//2)]
        else:
            input_2 = input_data
        data_2 = cv2.resize(input_2, (image_width>>1, image_height>>1), interpolation=interpolation)
        if display_timing:
            print(  f' === ViewerImage.reduce_half(): OpenCV resize from {input_data.shape} to '
                    f'{data_2.shape} --> {int((get_time()-start_0)*1000)} ms')
        return data_2

    # Seems interesting to have this method, but it is not so obvious
    def get_data_for_ratio(self, ratio, display_timing=False):
        prev_shape = self._data.shape
        image_height, image_width  = self._data.shape[:2]
        downscale_interpolation = cv2.INTER_AREA
        # if ratio is >2, start with integer downsize which is much faster
        # we could add this condition opencv_downscale_interpolation==cv2.INTER_AREA
        if ratio<=0.5:
            data_2 = self.reduce_half(self._data, interpolation=downscale_interpolation, display_timing=display_timing)
            if ratio<=0.25:
                data_4 = self.reduce_half(data_2, interpolation=downscale_interpolation, display_timing=display_timing)
                return data_4
            else:
                return data_2
        else:
            return self._data

    def set_filename(self, fn):
        self.filename = fn

    def __sizeof__(self):
        # approximative estimation
        size = self.data.nbytes
        for v in vars(self):
            # print(f" v {v} {self.__dict__[v].__sizeof__()}")
            size += self.__dict__[v].__sizeof__()
        return size



from qimview.utils.viewer_image import *
from qimview.utils.utils import get_time
import cv2
import numpy as np


def read_opencv(image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False):
    if verbose:
        open_cv2_start = get_time()
    cv_size = {'full': cv2.IMREAD_COLOR,
                '1/2': cv2.IMREAD_REDUCED_COLOR_2,
                '1/4': cv2.IMREAD_REDUCED_COLOR_4,
                '1/8': cv2.IMREAD_REDUCED_COLOR_8,
                }
    flags = cv_size[read_size]
    if image_buffer is None:
        cv2_im = cv2.imread(image_filename, flags)
    else:
        bytes_as_np_array = np.frombuffer(image_buffer, dtype=np.uint8)
        cv2_im = cv2.imdecode(bytes_as_np_array, flags) 

    open_cv2_start2 = get_time()
    # transform default opencv BGR to RGB
    if use_RGB:
        im = cv2.cvtColor(cv2_im, cv2.COLOR_BGR2RGB)

    if verbose:
        last_time = get_time()
        print(" cv2 imread {0} took {1:0.3f} ( {2:0.3f} + {3:0.3f} ) sec.".format(image_filename, last_time - open_cv2_start,
                                                                    open_cv2_start2 - open_cv2_start,
                                                                    last_time - open_cv2_start2),
                )
    downscale = {'full': 1, '1/2': 2, '1/4': 4, '1/8': 8}[read_size]
    viewer_image = ViewerImage(im, precision=8, downscale=downscale, channels=ImageFormat.CH_RGB if use_RGB else ImageFormat.CH_BGR)
    return viewer_image


def opencv_supported_formats():
    # according to the doc, opencv supports the following formats
    # Windows bitmaps - *.bmp, *.dib (always supported)
    # JPEG files - *.jpeg, *.jpg, *.jpe (see the Note section)
    # JPEG 2000 files - *.jp2 (see the Note section)
    # Portable Network Graphics - *.png (see the Note section)
    # WebP - *.webp (see the Note section)
    # Portable image format - *.pbm, *.pgm, *.ppm *.pxm, *.pnm (always supported)
    # PFM files - *.pfm (see the Note section)
    # Sun rasters - *.sr, *.ras (always supported)
    # TIFF files - *.tiff, *.tif (see the Note section)
    # OpenEXR Image files - *.exr (see the Note section)
    # Radiance HDR - *.hdr, *.pic (always supported)
    # Raster and Vector geospatial data supported by GDAL (see the Note section)    
    return [ '.bmp', '.dib', '.jpeg', '.jpg', '.jpe', '.jp2', '.png', '.webp', '.pbm', '.ppm', '.pxm', 
            '.pnm', '.pfm', '.sr', '.ras', '.tiff', '.tif', '.exr', '.hdr', '.pic']





from utils.qt_imports import QtWidgets
import sys
from video_player.video_player import VideoPlayer


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = VideoPlayer(app, open_button=True)
    sys.exit(app.exec_())



from ..utils.qt_imports import QtWidgets, QtCore

class NumericParameterGui(QtWidgets.QSlider):
    """
    For the moment, it can only be a slider with associated text
    """
    def __init__(self, name, param, callback, layout=None, parent_name=""):
        QtWidgets.QSlider.__init__(self, QtCore.Qt.Horizontal)
        self.name = name
        self.param = param
        self.callback = callback
        self.event_recorder = None
        self.parent_name = parent_name
        self.widget_name = f"slider_{self.parent_name}_{self.name}"
        self.created = False
        self.decimals = 2
        if layout is not None:
            self.create()
            self.add_to_layout(layout)

    def set_event_recorder(self, evtrec):
        self.event_recorder = evtrec
        if self.event_recorder is not None:
            self.event_recorder.register_widget(id(self), self.widget_name)

    def register_event_player(self, event_player):
        event_player.register_widget(self.widget_name, self)

    def create(self, moved_callback=False):
        self.label = QtWidgets.QLabel("")
        self.setRange(self.param.range[0], self.param.range[1])
        self.setValue(self.param.value)
        self.changed()
        if moved_callback:
            self.sliderMoved.connect(lambda: self.changed(self.callback))
        else:
            self.valueChanged.connect(lambda: self.changed(self.callback))
        self.created = True

    def add_to_layout(self, layout):
        if not self.created:
            self.create()
        layout.addWidget(self.label)
        layout.addWidget(self)

    def reset(self):
        self.setValue(self.param.default_value)

    def changed(self, callback=None):
        self.param.int = int(self.value())
        self.label.setText(f"{self.name} {self.param.float:0.{self.decimals}f}")
        if callback is not None:
            callback()

    def mouseDoubleClickEvent(self, evt):
        self.reset()

    def event(self, evt):
        if self.event_recorder is not None:
            if evt.spontaneous():
                self.event_recorder.store_event(self, evt)
        return QtWidgets.QSlider.event(self, evt)




from qimview.parameters.numeric_parameter import NumericParameter

class ImageFilterParameters:
    def __init__(self):
        # white/black levels
        # default_black = int(4095*5/100)
        default_black = 0
        self.black_level = NumericParameter(default_black, default_black, [0, 4095]  , 4095)
        self.white_level = NumericParameter(4095, 4095,            [480, 4095], 4095)
        # gamma curve coefficient
        self.gamma       = NumericParameter(100, 100,            [50, 300], 100)
        # white balance coefficients
        self.g_b = NumericParameter(256, 256, [50, 512], 256)
        self.g_r = NumericParameter(256, 256, [50, 512], 256)
        # Saturation
        self.saturation = NumericParameter(50, 50, [0, 150], 50)
        # Image difference factor
        self.imdiff_factor = NumericParameter(30, 30, [1, 100], 10)

    def copy_from(self, p):
        for v in vars(self):
            if isinstance(self.__dict__[v], NumericParameter):
                self.__dict__[v].copy_from(p.__dict__[v])

    def is_equal(self, other):
        if not isinstance(other, ImageFilterParameters):
            return NotImplemented
        for v in vars(self):
            var1 = self.__dict__[v]
            if isinstance(var1, NumericParameter):
                var2 = other.__dict__[v]
                if var1.float != var2.float:
                    return False
        return True

    def __repr__(self):
        return f"<ImageFilterParameters {id(self)}>"

    def __str__(self):
        res = ""
        for v in vars(self):
            res += f"{v}:{self.__dict__[v]}; "
        return res




from ..utils.utils import get_time
import json
from .qtdump import *

class EventRecorder:

    def __init__(self, filename=None):
        self.event_list = []
        self.event_start = get_time()
        self.filename = filename
        self.widgets = dict()
        self.verbose = True

    def register_widget(self, id, name):
        self.widgets[id] = name

    def store_event(self, widget, evt):
        # event_filter = MOUSE_EVENTS + RESIZE_EVENTS + WHEEL_EVENT
        if evt.type() in MOUSE_EVENTS and evt.spontaneous():
            evt_info = QtDump.mouseevent2dict(evt)
        else:
            if evt.type() in RESIZE_EVENTS and evt.spontaneous():
                evt_info = QtDump.resizeevent2dict(evt)
            else:
                if evt.type() in WHEEL_EVENT and evt.spontaneous():
                    evt_info = QtDump.wheelevent2dict(evt)
                else:
                    return
        evt_time = get_time() - self.event_start
        evt_info = {'info': evt_info, 'time': evt_time}
        if id(widget) in self.widgets:
            evt_info['widget_name'] = self.widgets[id(widget)]
        if self.verbose:
            print(f"recording {evt_info}")
        self.event_list.append(evt_info)

    def save_screen(self, widget):
        image_hash = QtDump.get_screen_hash(widget)
        print(f"image hash ${image_hash}")
        evt_time = get_time() - self.event_start
        evt_info = {'info': {'type':'save_screen', 'hash':image_hash}, 'time': evt_time}
        if id(widget) in self.widgets:
            evt_info['widget_name'] = self.widgets[id(widget)]
        self.event_list.append(evt_info)

    def save_events(self, filename=None):
        if filename is None:
            filename = self.filename
        if filename is not None:
            with open(filename, 'w') as fp:
                json.dump({'events': self.event_list}, fp)
        else:
            print("Error: EventRecorder.save_events() no filename given !")



#!/bin/python

import sys
import argparse
import os
import glob

from qimview.utils.qt_imports import QtWidgets, QtCore
from qimview.image_viewers    import MultiView, ViewerType

def get_filenames():
    filenames = []
    # Ask for input file
    selected_files =  QtWidgets.QFileDialog.getOpenFileNames(caption="miview: Select  one or various input images")
    filenames.extend(selected_files[0])
    return filenames


def main():
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('images', nargs='*', help='input images')
    parser.add_argument('-v', '--viewer', type=str, choices={'gl', 'qt', 'shader'}, default='qt',
                        help="Viewer mode, qt: standard qt display, gl: use opengl,  shader: enable opengl with "
                             "shaders")
    parser.add_argument('-l', '--layout', type=str, default='0', help='Set the layout (number of images in comparison on the window), if 0 try to use the number of input images')

    args = parser.parse_args()
    _params = vars(args)

    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationDisplayName('mview ' + ' '.join(sys.argv[1:]))

    filenames = []
    for im in _params['images']:
        filenames.extend(glob.glob(im,recursive=False))
    checked_filenames = []
    for f in filenames:
        if os.path.isfile(f): checked_filenames.append(f)
        else: print(f"Filename {f} not found")

    if len(checked_filenames) == 0:
        # Ask for input file(s)
        checked_filenames = get_filenames()

    mode = {
        'qt': ViewerType.QT_VIEWER,
        'gl': ViewerType.OPENGL_VIEWER,
        'shader': ViewerType.OPENGL_SHADERS_VIEWER
    }[_params['viewer']]

    # Create a simple widget and layout to put the multiview inside,
    # so that toggleFullscreen can work
    main_window = QtWidgets.QMainWindow()
    main_widget = QtWidgets.QWidget(main_window)
    main_window.setCentralWidget(main_widget)    
    multiview_layout = QtWidgets.QVBoxLayout()
    main_widget.setLayout(multiview_layout)
    mv = MultiView(parent=main_widget, viewer_mode=mode)
    multiview_layout.addWidget(mv, 1)

    # table_win.setWindowTitle('Image Set Comparison ' + title_string)
    # table_win.set_default_report_file(default_report_file + '.json')
    # table_win.CreateImageDisplay(image_list)
    def get_name(path, maxlength=20):
        return os.path.splitext(os.path.basename(path))[0][-maxlength:]

    images_dict = {}

    for idx,im in enumerate(checked_filenames):
        images_dict[f"{idx}_{get_name(im)}"] = im
    mv.set_images(images_dict)
    mv.update_layout()
    # table_win.resize(3000, 1800)

    nb_inputs = len(checked_filenames)
    mv.show()
    if nb_inputs>=1 and nb_inputs<=9:
        mv.set_number_of_viewers(nb_inputs)
        mv.viewer_grid_layout.update()
        mv.update_image()
        mv.setFocus()

    main_window.show()
    main_window.resize(1000, 800)

    app.exec()

if __name__ == '__main__':
    main()



from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup
# import os
# import distutils
# distutils.log.set_verbosity(1)

# from distutils.command.build_ext import build_ext
# compiler/linker args depending on the platform: check how to set it up
# compiler_args = ['-std=c++11', '-fopenmp', '-O3']
# compiler_args = ['/openmp', '/O2', '/arch:AVX512']
# compiler_args = ['-std=c++11', '-stdlib=libc++', '-mmacosx-version-min=10.7', '/openmp', '/Ox']

class build_ext_subclass( build_ext ):
    # win32, linux, darwin
    # msvc, unix
    copt =  {
        'win32' : ['/openmp',  '/O2', ],  # , '/fp:fast','/favor:INTEL64','/Og'],
        'linux' : ['-fopenmp', '-O3', '-std=c++17', ],
        'darwin': [            '-O3', '-std=c++17', ],
    }
    lopt =  {
        'darwin' : ['-L/usr/local/opt/libomp/lib'],
        'linux' : ['-fopenmp', ],
    }

    def build_extensions(self):
        # c = self.compiler.compiler_type
        # print(f"compiler_type {c}")
        import sys
        print(f" *** platform {sys.platform}")
        c = sys.platform
        if c in self.copt:
           for e in self.extensions:
               e.extra_compile_args = self.copt[ c ]
        if c in self.lopt:
            for e in self.extensions:
                e.extra_link_args = self.lopt[ c ]
        build_ext.build_extensions(self)

setup_args = dict(
    ext_modules = [
        Pybind11Extension("qimview_cpp",
            ["qimview/cpp_bind/qimview_cpp.cpp"],
            depends = [ 
                'qimview/cpp_bind/image_histogram.hpp',
                'qimview/cpp_bind/image_resize.hpp',
                'qimview/cpp_bind/image_to_rgb.hpp',
                ],
            # Example: passing in the version to the compiled code
            # define_macros = [('VERSION_INFO', __version__)],
            include_dirs=['qimview/cpp_bind','/usr/local/opt/libomp/include'],
            ),
    ],
    cmdclass = {'build_ext': build_ext_subclass } 
)
setup(**setup_args)



import qimview_cpp
import numpy as np
import cv2
import timeit 

w, h   = 5000, 4000
w2, h2 = 2500, 2000

# w, h   = 4, 2
# w2, h2 = 2, 1

a = (np.random.rand(h,w,3)*256).astype(np.uint8)
b = np.zeros((h2,w2,a.shape[2]), dtype=np.uint8)


# %timeit qimview_cpp.image_binning_2x2(a,b)
# %timeit c = cv2.resize(a, (w2, h2), interpolation=cv2.INTER_AREA)

qimview_cpp.image_binning_2x2_test1(a,b)
c = cv2.resize(a, (w2, h2), interpolation=cv2.INTER_AREA)

print(f" diff min {np.min(c-b)} max {np.max(c-b)}")



from qimview.utils.utils      import deep_getsizeof
from qimview.utils.thread_pool import ThreadPool
from qimview.utils.qt_imports import QtWidgets

from collections import deque
from typing import List, TypeVar, Optional, Generic, Any, Tuple

TId    = TypeVar("TId")
TValue = TypeVar("TValue")
TExtra = TypeVar("TExtra")

class BaseCache(Generic[TId, TValue, TExtra]):
    # --- Private methods
    def __init__(self, name : str =""):
        # Python list and deque are thread-safe
        self.cache      : deque[Tuple[TId, TValue, TExtra]] = deque()
        self.cache_list : List[TId]                                   = []
        self.cache_size : int                                         = 0
        # Max size in Mb
        self.max_cache_size : int                              = 2000
        self.verbose        : bool                             = False
        self.cache_unit     : int                              = 1024*1024 # Megabyte
        self.thread_pool    : ThreadPool                       = ThreadPool()
        self.memory_bar     : Optional[QtWidgets.QProgressBar] = None
        self._name          : str                              = name
        # Avoid changing progressbar inside a thread so no need for mutex

    # --- Protected methods
    def _print_log(self, message : str) -> None:
        if self.verbose:
            print(message)

    # --- Public methods
    def set_memory_bar(self, progress_bar:QtWidgets.QProgressBar) -> None:
        self.memory_bar = progress_bar
        self.memory_bar.setRange(0, self.max_cache_size)
        self.memory_bar.setFormat("%v Mb")

    def reset(self) -> None:
        self.cache = deque()
        self.cache_list = []
        self.cache_size = 0

    def set_max_cache_size(self, size : int) -> None:
        self.max_cache_size = size
        self.check_size_limit()
        if self.memory_bar is not None:
            self.memory_bar.setRange(0, self.max_cache_size)

    def search(self, id : TId) -> Optional[Tuple[TId, TValue,TExtra]]:
        res = None
        if id in self.cache_list:
            pos = self.cache_list.index(id)
            # print(f"pos {pos} len(cache) {len(self.cache)}")
            try:
                res = self.cache[pos]
            except Exception as e:
                print(f" Error in getting cache data: {e}")
                self._print_log(f" *** Cache {self._name}: search() cache_list {len(self.cache_list)} cache {len(self.cache)}")
                res = None
        return res

    def append(self, id : TId, value: TValue, extra: TExtra, check_size=True) -> None:
        """
        :param id: cache element identifier
        :param value: cache value, typically numpy array of the image
        :param extra: additional data in the cache
        :return:
        """
        # update cache
        self._print_log(f"added size {deep_getsizeof([id, value, extra], set())}")
        self.cache.append((id, value, extra))
        self.cache_list.append(id)
        self._print_log(f" *** Cache {self._name}: append() cache_list {len(self.cache_list)} cache {len(self.cache)}")
        if check_size:
            self.check_size_limit()
    
    def remove(self, id:TId) -> bool:
        """ Remove id from cache
            returns: True if removed False otherwise (not found)
        """ 
        res = self.search(id)
        if res is None: return False
        self.cache.remove(res)
        self.cache_list.remove(id)
        return True

    def get_cache_size(self) -> int:
        size = deep_getsizeof(self.cache, set())
        return size
    
    def update_progress(self):
        if self.memory_bar is not None:
            new_progress_value = int(self.cache_size/self.cache_unit+0.5)
            if new_progress_value != self.memory_bar.value():
                self.memory_bar.setValue(new_progress_value)

    def check_size_limit(self, update_progress : bool = False) -> None:
        self._print_log(" *** Cache: check_size_limit()")
        cache_size = self.get_cache_size()
        while cache_size >= self.max_cache_size * self.cache_unit:
            self.cache.popleft()
            self.cache_list.pop(0)
            self._print_log(" *** Cache: pop ")
            cache_size = self.get_cache_size()
        self._print_log(f" *** Cache::append() {self._name} {cache_size/self.cache_unit} Mb; size {len(self.cache)}")
        self.cache_size = cache_size
        if update_progress:
            self.update_progress()



from qimview.utils.viewer_image import *
from qimview.utils.utils import get_time
import os
from typing import Optional
try:
    import simplejpeg
except:
    has_simplejpeg = False
else:
    has_simplejpeg = True

def read_jpeg_simplejpeg(image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False) -> Optional[ViewerImage]:
    # print(f"read_jpeg_simplejpeg use_RGB {use_RGB} running ...")
    if verbose:
        start_time = get_time()
    format = 'RGB' if use_RGB else 'BGR'
    try:
        if image_buffer is None:
            with open(image_filename, 'rb') as d:
                image_buffer = d.read()
    except IOError as e:
        print("read_jpeg_simplejpeg: I/O error({0}): {1}".format(e.errno, e.strerror))
    except Exception as e: #handle other exceptions such as attribute errors
        print(f"read_jpeg_simplejpeg: Unexpected error: {e}")
        return None

    try:
        im = simplejpeg.decode_jpeg(image_buffer, format)
    except Exception as e:
        print(f"read_jpeg_simplejpeg: Failed to decode jpeg {e}")
        return None

    if verbose:
        end_time = get_time()
        print(f" simplejpeg.decode_jpeg() {format} {os.path.basename(image_filename)} "
                f"took {int((end_time-start_time)*1000+0.5)} ms")
    viewer_image = ViewerImage(im, precision=8, downscale=1, channels=ImageFormat.CH_RGB if use_RGB else ImageFormat.CH_BGR)
    return viewer_image
